'use strict';

var css = require('css');

function getNodes(node) {
  var nodes = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];

  if (node.children) {
    var children = typeof node.children === 'function' ? node.children() : node.children;
    if (Array.isArray(children)) {
      children.forEach(function (child) {
        return getNodes(child, nodes);
      });
    }
  }

  if (typeof node === 'object') {
    nodes.push(node);
  }

  return nodes;
}

function markNodes(nodes) {
  nodes.forEach(function (node) {
    node.serializedWithJestGlamorReact = true;
  });
}

function getSelectors(nodes) {
  return nodes.reduce(function (selectors, node) {
    if (node instanceof global.HTMLElement) {
      return getSelectorsFromDOM(selectors, node);
    } else {
      return getSelectorsFromProps(selectors, node);
    }
  }, []);
}

// eslint-disable-next-line
function getSelectorsFromProps(selectors, node) {
  var props = typeof node.props === 'function' ? node.props() : node.props;
  var className = props.className || props.class;
  if (className) {
    selectors = selectors.concat(className.toString().split(' ').map(function (cn) {
      return `.${cn}`;
    }));
  }
  var dataProps = Object.keys(props).reduce(function (dProps, key) {
    if (key.startsWith('data-')) {
      dProps.push(`[${key}]`);
    }
    return dProps;
  }, []);
  if (dataProps.length) {
    selectors = selectors.concat(dataProps);
  }
  return selectors;
}

function getSelectorsFromDOM(selectors, node) {
  var root = document.createElement('div');
  root.appendChild(node);
  var allChildren = root.querySelectorAll('*');
  selectors = Array.from(allChildren).reduce(function (s, c) {
    s.push.apply(s, Array.from(c.classList).map(function (cn) {
      return `.${cn}`;
    }));
    s.push.apply(s, c.getAttributeNames().filter(function (key) {
      return key.startsWith('data-');
    }).map(function (key) {
      return `[${key}]`;
    }));
    return s;
  }, selectors);
  return selectors;
}

function filterChildSelector(baseSelector) {
  if (baseSelector.slice(-1) === '>') {
    return baseSelector.slice(0, -1);
  }
  return baseSelector;
}

function getAST(nodeSelectors, styleSheet) {
  var tags = typeof styleSheet === 'function' ? styleSheet().tags : styleSheet.tags;
  var styles = tags.map(function (tag) {
    return (/* istanbul ignore next */tag.textContent || ''
    );
  }).join('\n');
  var ast = css.parse(styles);
  var rules = ast.stylesheet.rules.filter(filter);
  var mediaQueries = getMediaQueries(ast, filter);

  ast.stylesheet.rules = [].concat(rules, mediaQueries);

  return ast;

  function filter(rule) {
    if (rule.type === 'rule') {
      return rule.selectors.some(function (selector) {
        var baseSelector = filterChildSelector(selector.split(/:| |\./).filter(function (s) {
          return !!s;
        })[0]);
        return nodeSelectors.some(function (sel) {
          return sel === baseSelector || sel === `.${baseSelector}`;
        });
      });
    }
    return false;
  }
}

function getStylesAndAllSelectors(nodeSelectors, styleSheet) {
  var ast = getAST(nodeSelectors, styleSheet);
  var allSelectors = ast.stylesheet.rules.reduce(function (s, r) {
    if (r.selectors) {
      s.push.apply(s, r.selectors);
    }
    return s;
  }, []);
  var styles = css.stringify(ast);
  return { styles, allSelectors };
}

function getMediaQueries(ast, filter) {
  return ast.stylesheet.rules.filter(function (rule) {
    return rule.type === 'media' || rule.type === 'supports';
  }).reduce(function (acc, mediaQuery) {
    mediaQuery.rules = mediaQuery.rules.filter(filter);

    if (mediaQuery.rules.length) {
      return acc.concat(mediaQuery);
    }

    return acc;
  }, []);
}

function shouldDive(node) {
  return typeof node.dive === 'function' && typeof node.type() !== 'string';
}

function isTagWithClassName(node) {
  return node.prop('className') && typeof node.type() === 'string';
}

function getClassNameFromEnzyme(received) {
  var tree = shouldDive(received) ? received.dive() : received;
  var components = tree.findWhere(isTagWithClassName);
  return components.length && components.first().prop('className');
}

function getClassNameFromTestRenderer(received) {
  return received.props.className || received.props.class;
}

function getClassNameFromCheerio(received) {
  /* istanbul ignore next */
  if (received[0].type === 'root') {
    // I'm honestly not sure we need this, but it was here
    // and I don't want to break it...
    return received.children().first().attr('class');
  } else {
    return received.attr('class');
  }
}

// eslint-disable-next-line complexity
function getClassNames(received) {
  var className = void 0;

  if (received) {
    if (received instanceof HTMLElement) {
      className = received.className;
    } else if (received.$$typeof === Symbol.for('react.test.json')) {
      className = getClassNameFromTestRenderer(received);
    } else if (typeof received.findWhere === 'function') {
      className = getClassNameFromEnzyme(received);
    } else if (received.cheerio === '[cheerio object]') {
      className = getClassNameFromCheerio(received);
    }
  }

  return className ? className.split(/\s/) : [];
}

module.exports = {
  getNodes,
  markNodes,
  getSelectors,
  getAST,
  getStylesAndAllSelectors,
  getClassNames
};